@@ -1,4 +1,6 @@
 from .functions import get_spark, get_dbutils, data_convert
+from .constants import PANDAS_TYPES, PYSPARK_TYPES, KOALAS_TYPES, PANDAS_ON_SPARK_TYPES, PANDAS, KOALAS,\ 
+SPARK, PANDAS_ON_SPARK, IN_PANDAS_ON_SPARK
 
 # Azure
 from azure.cosmos import CosmosClient, PartitionKey
@@ -34,17 +36,17 @@ from random import choice
 from json import loads
 
 # TABLE
-def read_table(table_name: str, db: str = "default", as_type: str = "koalas", index: Union[str, List[str], None] = None, engine: str = "koalas", threads: int = 2):
-    def read_table_f(table_name: str, db: str = "default", as_type: str = "koalas", index: Union[str, List[str], None] = None, engine: str = "koalas"):
-        if engine == "koalas":
+def read_table(table_name: str, db: str = "default", as_type: str = KOALAS, index: Union[str, List[str], None] = None, engine: str = KOALAS, threads: int = 2):
+    def read_table_f(table_name: str, db: str = "default", as_type: str = KOALAS, index: Union[str, List[str], None] = None, engine: str = KOALAS):
+        if engine == KOALAS:
             return data_convert(koalas_read_table(f"{db}.{table_name}", index_col = index), as_type=as_type)
-        elif engine == "spark":
+        elif engine == PYSPARK:
             query = f"SELECT * FROM {db}.{table_name}"
             if index:
                 return data_convert(spark.sql(query), as_type=as_type).set_index(index)
             else:
                 return data_convert(spark.sql(query), as_type=as_type)
-        elif engine == "pandas":
+        elif engine == PANDAS:
             # Pandas doesn't have this originally, we use ps for this.
             # Remove .as_pandas when data_convert supports ps.
             return data_convert(ps_read_table(f"{db}.{table_name}", index_col = index).to_pandas(), as_type=as_type)
@@ -66,7 +68,7 @@ def write_table(df, table_name, db: str = "default", delta_path = "/FileStore/ta
     def write_table_f(df, table_name, db: str = "default", delta_path = "/FileStore/tables/", mode: str = "overwrite", partitions: Union[str, List[str], None] = None, index: Union[str, List[str], None] = None) -> None:
         # Write .delta file in datalake
         delta_path = f'{delta_path}/{db}/{table_name}/'.replace('//', '/')
-        data_convert(df, as_type="koalas").to_delta(delta_path, mode=mode, partition_cols = partitions, index_col = index)
+        data_convert(df, as_type=KOALAS).to_delta(delta_path, mode=mode, partition_cols = partitions, index_col = index)
         # Create DB if not exist
         spark.sql(f"CREATE DATABASE IF NOT EXISTS {db}")
         # Create Table
@@ -114,9 +116,9 @@ def write_table(df, table_name, db: str = "default", delta_path = "/FileStore/ta
     return q.join()
 
 # CSV
-def read_csv(file_name: str, path: str, sep: str = ",", header: bool = 0, source: str = "dbfs", as_type: str = "koalas", engine: str = "koalas", threads: int = 2):
-    def read_csv_f(file_name: str, path: str, sep: str = ",", header: bool = 0, source: str = "dbfs", as_type: str = "koalas", engine: str = "koalas"):
-        if engine == "koalas":
+def read_csv(file_name: str, path: str, sep: str = ",", header: bool = 0, source: str = "dbfs", as_type: str = KOALAS, engine: str = KOALAS, threads: int = 2):
+    def read_csv_f(file_name: str, path: str, sep: str = ",", header: bool = 0, source: str = "dbfs", as_type: str = KOALAS, engine: str = KOALAS):
+        if engine == KOALAS:
             if header and isinstance(header, bool):
                 header = 'infer'
             elif not header and isinstance(header, bool):
@@ -131,7 +133,7 @@ def read_csv(file_name: str, path: str, sep: str = ",", header: bool = 0, source
                 for df_column in df.columns:
                     df = df.rename(columns = {df_column: f"{df.loc[header_index][df_column]}"})
                 df = df.loc[header_index:]
-        elif engine == "pandas":
+        elif engine == PANDAS:
             if header and isinstance(header, bool):
                 header = 'infer'
             elif not header and isinstance(header, bool):
@@ -140,7 +142,7 @@ def read_csv(file_name: str, path: str, sep: str = ",", header: bool = 0, source
                 header = header
 
             df = pandas_read_csv(f"/{source}{path}/{file_name}", sep = sep, header = header)
-        elif engine == "spark":
+        elif engine == PYSPARK:
             if isinstance(header, int): 
                 header_index = header
                 header = False
@@ -172,7 +174,7 @@ def write_csv(df, file_name: str, path: str, sep: str = ',', nas: str = '', head
     del kwds['threads']
     def write_csv_f(df, file_name: str, path: str, sep: str = ',', nas: str = '', header: bool = True, mode: str = 'overwrite', num_files: int = 1):
         _temp = f"temp_{file_name.split('.')[0]}"
-        data_convert(df, as_type="koalas").to_csv(path=fr'{path}/{_temp}/', sep = sep, na_rep = nas, header = header, mode = mode, num_files= num_files)
+        data_convert(df, as_type=KOALAS).to_csv(path=fr'{path}/{_temp}/', sep = sep, na_rep = nas, header = header, mode = mode, num_files= num_files)
         files = [file.name for file in dbutils.fs.ls(f"{path}/{_temp}/") if not file.name.startswith("_")]
         i = 0
         for file in files:
@@ -199,11 +201,11 @@ def write_csv(df, file_name: str, path: str, sep: str = ',', nas: str = '', head
     return list(map(lambda kwd: pool.apply_async(write_csv_f, kwds=kwd).get(), kwds))
 
 # PARQUET
-def read_parquet(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", engine: str = "koalas", threads: int = 2) -> Union[DataFrame, List[DataFrame]]:
-    def read_parquet_f(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", engine: str = "koalas"):
-        if engine.lower() == "koalas":
+def read_parquet(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, engine: str = KOALAS, threads: int = 2) -> Union[DataFrame, List[DataFrame]]:
+    def read_parquet_f(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, engine: str = KOALAS):
+        if engine.lower() == KOALAS:
             return data_convert(koalas_read_parquet(f"{source}:{path}/{file_name}"), as_type=as_type)
-        elif engine.lower() == "spark":
+        elif engine.lower() == PYSPARK:
             try:
                 return data_convert(spark.read.format("parquet").load(f"{source}:{path}/{file_name}"), as_type=as_type)
             except:
@@ -211,7 +213,7 @@ def read_parquet(file_name: str, path: str, source: str = "dbfs", as_type: str =
                     return data_convert(spark.read.format("parquet").load(f"{source}:{path}/{file_name}/"), as_type=as_type)
                 except:
                     raise
-        elif engine.lower() == "pandas":
+        elif engine.lower() == PANDAS:
             # Pandas has an engine option that can be pyarrow or fastparquet. Defaults to pyarrow.
             return data_convert(pandas_read_parquet(f"/{source}/{path}/{file_name}"), as_type=as_type)
         
@@ -228,7 +230,7 @@ def write_parquet(df, file_name: str, path: str, mode: str = "overwrite", num_fi
     del kwds['threads']
     def write_parquet_f(df, file_name: str, path: str, mode: str = "overwrite", num_files: int = 1):
         _temp = f"temp_{file_name.split('.')[0]}"
-        data_convert(df, as_type="koalas").to_parquet(path=fr'{path}/{_temp}/', mode = mode, num_files= num_files)
+        data_convert(df, as_type=KOALAS).to_parquet(path=fr'{path}/{_temp}/', mode = mode, num_files= num_files)
         files = [file.name for file in dbutils.fs.ls(f"{path}/{_temp}/") if not file.name.startswith("_")]
         i = 0
         for file in files:
@@ -255,9 +257,9 @@ def write_parquet(df, file_name: str, path: str, mode: str = "overwrite", num_fi
     return list(map(lambda kwd: pool.apply_async(write_parquet_f, kwds=kwd).get(), kwds))
 
 # JSON
-def read_json(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", engine: str = "koalas", threads: int = 2):
-    def read_json_f(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", engine: str = "koalas"):
-        if engine == "koalas":
+def read_json(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, engine: str = KOALAS, threads: int = 2):
+    def read_json_f(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, engine: str = KOALAS):
+        if engine == KOALAS:
             try:
                 return data_convert(koalas_read_json(f"{source}:{path}/{file_name}", lines = True), as_type=as_type)
             except:
@@ -265,7 +267,7 @@ def read_json(file_name: str, path: str, source: str = "dbfs", as_type: str = "k
                     return data_convert(koalas_read_json(f"{source}:{path}/{file_name}", lines = False), as_type=as_type)
                 except:
                     raise
-        elif engine == "spark":
+        elif engine == PYSPARK:
             try:
                 return data_convert(spark.read.format("json").load(f"{source}:{path}/{file_name}"), as_type=as_type)
             except:
@@ -273,7 +275,7 @@ def read_json(file_name: str, path: str, source: str = "dbfs", as_type: str = "k
                     return data_convert(spark.read.format("json").load(f"{source}:{path}/{file_name}/"), as_type=as_type)
                 except:
                     raise
-        elif engine == "pandas":
+        elif engine == PANDAS:
             try:
                 return data_convert(pandas_read_json(f"/{source}/{path}/{file_name}", lines = True), as_type=as_type)
             except:
@@ -290,7 +292,7 @@ def read_json(file_name: str, path: str, source: str = "dbfs", as_type: str = "k
 
 def write_json(df, file_name: str, path: str, num_files: int = 1):
     _temp = f"temp_{file_name.split('.')[0]}"
-    data_convert(df, as_type="koalas").to_json(path=fr'{path}/{_temp}/', num_files=num_files)
+    data_convert(df, as_type=KOALAS).to_json(path=fr'{path}/{_temp}/', num_files=num_files)
     files = [file.name for file in dbutils.fs.ls(f"{path}/{_temp}/") if not file.name.startswith("_")]
     i = 0
     for file in files:
@@ -305,11 +307,11 @@ def write_json(df, file_name: str, path: str, num_files: int = 1):
         dbutils.fs.rm(f"{path}/{_temp}", True)
 
 # EXCEL
-def read_excel(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", header: bool = True, engine: str = "koalas", threads: int = 2):
-    def read_excel_f(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", header: bool = True, engine: str = "koalas"):
-        if engine == "koalas":
+def read_excel(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, header: bool = True, engine: str = KOALAS, threads: int = 2):
+    def read_excel_f(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, header: bool = True, engine: str = KOALAS):
+        if engine == KOALAS:
             return data_convert(koalas_read_excel(f"{source}:{path}/{file_name}"), as_type=as_type)
-        elif engine == "spark":
+        elif engine == PYSPARK:
             try:
                 return data_convert(spark.read.format("com.crealytics.spark.excel").option("location", f"{path}/{file_name}").option("header", header).option("inferSchema", True).load(f"{path}/{file_name}"), as_type=as_type)
             except:
@@ -318,7 +320,7 @@ def read_excel(file_name: str, path: str, source: str = "dbfs", as_type: str = "
                 except Py4JJavaError as e:
                     if "java.lang.ClassNotFoundException:" in str(e):
                         raise Exception("Maven dependency not installed in cluster. Install com.crealytics.spark.excel.")
-        elif engine == "pandas":
+        elif engine == PANDAS:
             return data_convert(pandas_read_excel(f"/{source}/{path}/{file_name}"), as_type=as_type)
         
     if isinstance(file_name, list):
@@ -366,7 +368,7 @@ def write_excel(df, file_name: str, path: str, source: str = "dbfs", extension:
         try:
             with ExcelWriter(fr"{file_name}") as writer:
                 for dfs in df:
-                    data_convert(dfs, as_type = "koalas").to_excel(writer, sheet_name=sheet_name[sheet_pos])
+                    data_convert(dfs, as_type = KOALAS).to_excel(writer, sheet_name=sheet_name[sheet_pos])
                     sheet_pos = sheet_pos + 1
             system(fr"mv /databricks/driver/{file_name} /{source}/{path}/{file_name}")
         except:
@@ -396,26 +398,26 @@ def read_delta(file_name: str,
                source: str = "dbfs", 
                version: int = None, 
                timestamp: str = None, 
-               as_type: str = "koalas",
-               engine: str  = "spark",
+               as_type: str = KOALAS,
+               engine: str  = PYSPARK,
                threads: int = 2):
     def read_delta_f(file_name: str, 
                    path: str, 
                    source: str = "dbfs", 
                    version: int = None, 
                    timestamp: str = None, 
-                   as_type: str = "koalas",
-                   engine: str  = "koalas"):
-        if engine == "koalas":
+                   as_type: str = KOALAS,
+                   engine: str  = KOALAS):
+        if engine == KOALAS:
             return data_convert(koalas_read_delta(f"{source}:{path}/{file_name}", version=version, timestamp=timestamp), as_type=as_type)
-        elif engine == "spark":
+        elif engine == PYSPARK:
             if version:
                 return data_convert(spark.read.format("delta").option("versionAsOf", version).load(f"{source}:{path}/{file_name}"), as_type = as_type)
             elif timestamp:
                 return data_convert(spark.read.format("delta").option('timestampAsOf', timestamp).load(f"{source}:{path}/{file_name}"), as_type = as_type)
             else:
                 return data_convert(spark.read.format("delta").load(f"{source}:{path}/{file_name}"), as_type = as_type)
-        elif engine == "pandas":
+        elif engine == PANDAS:
             # Pandas doesn't have this originally, we use ps for this.
             # Remove .as_pandas when data_convert supports ps.
             return data_convert(ps_read_delta(f"/{source}/{path}/{file_name}").to_pandas(), as_type = as_type)
@@ -431,7 +433,7 @@ def write_delta(df, file_name: str, path: str, source: str = "dbfs", mode = "ove
     def write_delta_f(df, file_name: str, path: str, source: str = "dbfs", mode = "overwrite"):
         delta_path = path
         _temp = file_name#f"temp_{file_name.split('.')[0]}"
-        data_convert(df, as_type="koalas").to_delta(path=fr'{path}/{_temp}/', mode = mode)
+        data_convert(df, as_type=KOALAS).to_delta(path=fr'{path}/{_temp}/', mode = mode)
         files = [file.name for file in dbutils.fs.ls(f"{path}/{_temp}/") if not file.name.startswith("_")]
         i = 0
         for file in files:
@@ -459,13 +461,13 @@ def write_delta(df, file_name: str, path: str, source: str = "dbfs", mode = "ove
     return list(map(lambda kwd: pool.apply_async(write_delta_f, kwds=kwd).get(), kwds))
 
 # ORC
-def read_orc(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", engine: str = "koalas", threads: int = 2):
-    def read_orc_f(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", engine: str = "koalas"):
-        if engine == "koalas":
+def read_orc(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, engine: str = KOALAS, threads: int = 2):
+    def read_orc_f(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, engine: str = KOALAS):
+        if engine == KOALAS:
             return data_convert(koalas_read_orc(f"{source}:{path}/{file_name}"), as_type=as_type)
-        elif engine == "spark":
+        elif engine == PYSPARK:
             return data_convert(spark.read.format("orc").load(f"{source}:{path}/{file_name}"), as_type = as_type)
-        elif engine == "pandas":
+        elif engine == PANDAS:
             # Remove ps.to_pandas() when data_convert can support ps convertions.
             return data_convert(ps_read_orc(f"{source}:{path}/{file_name}").to_pandas(), as_type = as_type)
             # Current pyarrow version has a bug, pyspark.pandas is recommended.
@@ -486,7 +488,7 @@ def write_orc(df, file_name: str, path: str, source: str = "dbfs", mode = "overw
     del kwds['threads']
     def write_orc_f(df, file_name: str, path: str, source: str = "dbfs", mode = "overwrite", num_files: int = 1):
         _temp = file_name#f"temp_{file_name.split('.')[0]}"
-        data_convert(df, as_type="koalas").to_orc(path=fr'{path}/{_temp}/', mode = mode, num_files = num_files)
+        data_convert(df, as_type=KOALAS).to_orc(path=fr'{path}/{_temp}/', mode = mode, num_files = num_files)
         files = [file.name for file in dbutils.fs.ls(f"{path}/{_temp}/") if not file.name.startswith("_")]
         i = 0
         for file in files:
@@ -514,8 +516,8 @@ def write_orc(df, file_name: str, path: str, source: str = "dbfs", mode = "overw
     return list(map(lambda kwd: pool.apply_async(write_orc_f, kwds=kwd).get(), kwds))
 
 # Cosmos
-def read_cosmos(endpoint: str, key: str, database: str, container: str, as_type: str = "koalas", engine: str = "pandas", threads: int = 2):
-    def read_cosmos_f(endpoint: str, key: str, database: str, container: str, as_type: str = "koalas", engine: str = "pandas"):
+def read_cosmos(endpoint: str, key: str, database: str, container: str, as_type: str = KOALAS, engine: str = PANDAS, threads: int = 2):
+    def read_cosmos_f(endpoint: str, key: str, database: str, container: str, as_type: str = KOALAS, engine: str = PANDAS):
         # CLIENT CONNECT
         client    = CosmosClient(endpoint, credential=key)
 
@@ -524,11 +526,11 @@ def read_cosmos(endpoint: str, key: str, database: str, container: str, as_type:
 
         # SELECT CONTAINER
         container = db.get_container_client(container)
-        if engine == "koalas":
+        if engine == KOALAS:
             raise NotImplementedError("Current version of PyArrow does not support this operation. Use Pandas as engine, transform all complex variables into str type and transform.")
-        elif engine == "spark":
+        elif engine == PYSPARK:
             raise NotImplementedError("Current version of PyArrow does not support this operation. Use Pandas as engine, transform all complex variables into str type and transform.")
-        elif engine == "pandas":
+        elif engine == PANDAS:
             return data_convert(DataFrame(container.read_all_items()), as_type=as_type)
     if isinstance(container, list):
         pool = ThreadPool(threads)
@@ -565,13 +567,13 @@ def write_cosmos(df, endpoint: str, key: str, database: str, container: str, uni
     return list(map(lambda item: pool.apply_async(store_data, kwds={"container": container, "request_body": item}), loads(df.to_json(orient='records'))))
 
 # SQL
-def read_sql(table_name: str, database: str, server: str, port: str, user: str, password: str, sql_type: str = "sqlserver", cert: str = ".database.windows.net", as_type: str = "koalas", engine = "koalas", threads: int = 2):
-    def read_sql_f(table_name: str, database: str, server: str, port: str, user: str, password: str, sql_type: str = "sqlserver", cert: str = ".database.windows.net", as_type: str = "koalas", engine = "koalas"):
-        if engine == "koalas":
+def read_sql(table_name: str, database: str, server: str, port: str, user: str, password: str, sql_type: str = "sqlserver", cert: str = ".database.windows.net", as_type: str = KOALAS, engine = KOALAS, threads: int = 2):
+    def read_sql_f(table_name: str, database: str, server: str, port: str, user: str, password: str, sql_type: str = "sqlserver", cert: str = ".database.windows.net", as_type: str = KOALAS, engine = KOALAS):
+        if engine == KOALAS:
             return data_convert(koalas_read_sql(table_name, con=f"jdbc:{sql_type}://{server}{cert}:{port};database={database};user={user}@{server};password={password}"), as_type=as_type)
-        elif engine == "spark":
+        elif engine == PYSPARK:
             raise NotImplementedError("Current version of Hivecode doesn't support Spark connection for SQL.")
-        elif engine == "pandas":
+        elif engine == PANDAS:
             # Remove to_pandas() after data_convert supports ps transformations.
             return data_convert(read_sql_query(f"SELECT * FROM {table_name}", f"jdbc:{sql_type}://{server}{cert}:{port};database={database};user={user}@{server};password={password}").to_pandas(), as_type = as_type)
     if isinstance(table_name, list):
@@ -580,13 +582,13 @@ def read_sql(table_name: str, database: str, server: str, port: str, user: str,
     elif isinstance(table_name, str):
         return read_sql_f(table_name, database, server, port, user, password, sql_type, cert, as_type, engine)
     
-def write_sql(df, table_name: str, database: str, server: str, port: str, user: str, password: str, sql_type: str = "sqlserver", cert: str = ".database.windows.net", engine = "koalas", threads: int = 2):
+def write_sql(df, table_name: str, database: str, server: str, port: str, user: str, password: str, sql_type: str = "sqlserver", cert: str = ".database.windows.net", engine = KOALAS, threads: int = 2):
     kwds = locals()
     del kwds['threads']
-    def write_sql_f(df, table_name: str, database: str, server: str, port: str, user: str, password: str, sql_type: str = "sqlserver", cert: str = ".database.windows.net", mode = "append", engine = "spark"):
-        if engine == "koalas":
+    def write_sql_f(df, table_name: str, database: str, server: str, port: str, user: str, password: str, sql_type: str = "sqlserver", cert: str = ".database.windows.net", mode = "append", engine = PYSPARK):
+        if engine == KOALAS:
             raise NotImplementedError("Current version of Hivecode doesn't support writing to a sqldb using Koalas.")
-        elif engine == "spark":
+        elif engine == PYSPARK:
             jdbcUrl = f"jdbc:sqlserver://{server}.database.windows.net:{port};database={database};"
             connectionProperties = {
               "user" : user,
@@ -595,7 +597,7 @@ def write_sql(df, table_name: str, database: str, server: str, port: str, user:
             }
 
             try:
-                data_convert(df, as_type="spark").write.jdbc(url=jdbcUrl, table=table_name, properties=connectionProperties, mode = mode)
+                data_convert(df, as_type=PYSPARK).write.jdbc(url=jdbcUrl, table=table_name, properties=connectionProperties, mode = mode)
             except Exception as e:
                 if "com.microsoft.sqlserver.jdbc.SQLServerException: The specified schema name" in str(e.java_exception):
                     # We create a new db and try again.
@@ -603,7 +605,7 @@ def write_sql(df, table_name: str, database: str, server: str, port: str, user:
                 elif "No suitable driver" == str(e.java_exception):
                     raise Exception("Driver not installed in cluster. Please install it using Maven Central 'com.microsoft.sqlserver.msi:msi-mssql-jdbc:2.0.3'")
                 raise
-        elif engine == "pandas":
+        elif engine == PANDAS:
             raise NotImplementedError("Current version of Hivecode doesn't support writing to a sqldb using Pandas.")
     
     # We encapsulate the data into a list if it's not already a list.
@@ -620,13 +622,13 @@ def write_sql(df, table_name: str, database: str, server: str, port: str, user:
 
 
 # AVRO
-def read_avro(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", engine: str = "spark", threads: int = 2):
-    def read_avro_f(file_name: str, path: str, source: str = "dbfs", as_type: str = "koalas", engine: str = "spark"):
-        if engine == "koalas":
+def read_avro(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, engine: str = PYSPARK, threads: int = 2):
+    def read_avro_f(file_name: str, path: str, source: str = "dbfs", as_type: str = KOALAS, engine: str = PYSPARK):
+        if engine == KOALAS:
             raise NotImplementedError("Current version of Koalas does not support this operation. Use Spark as engine.")
-        elif engine == "spark":
+        elif engine == PYSPARK:
             return data_convert(spark.read.format("avro").load(f"{source}:{path}/{file_name}"), as_type = as_type)
-        elif engine == "pandas":
+        elif engine == PANDAS:
             raise NotImplementedError("Current version of Pandas does not support this operation. Use Spark as engine.")
     if isinstance(file_name, list):
         for file in file_name:
